import numpy as np
from scipy.special import softmax

def invert_softmax(y, epsilon=1e-50):
    """
    Attempt to invert a softmax output to its original logits.
    This is a heuristic approach and doesn't guarantee the exact original logits.
    
    Parameters:
    - y: Softmax output array.
    - epsilon: A small value to replace zeros to avoid log(0).
    
    Returns:
    - Approximated original logits as a numpy array.
    """
    # Replace 0s with epsilon to avoid log(0)
    y = np.maximum(y, epsilon)
    
    # Apply the logarithm to approximate the original logits
    logits_approx = np.log(y)
    
    return logits_approx

        
def getInitalParameters(_self) -> np.ndarray:
    """Load default initial parameter vector based on function.
    
    Parameters:
        function: Name of the function to use.
    
    Returns:
        Initial parameter vector.
    """
    if _self.MCType == 'dtmc':
        x = np.array([np.random.uniform(low=b[0], high=b[1]) for b in bounds_x(_self)])
    elif _self.MCType == 'ihtmc':
        x = np.array([np.exp(np.random.uniform(low=b[0], high=b[1])) for b in bounds_x(_self)])
    else:
        raise ValueError('The type of Markov chain is not recognized.')
        
    # Initial state vector
    s0 = np.zeros(len(_self.states))
    s0[0] = 1.0
    
    return x,s0


def bounds_x(_self) -> list:
    """
    Determines the bounds for the parameter x based on the specified function.

    The bounds are determined based on the 'function' attribute of the class:
    - If 'function' is 'gompertz' or 'exponential', the bounds are set to (log(1E-20), log(50)).
    - For any other 'function' value, the bounds are set to (log(1E-18), log(50)).

    Returns:
        list: A tuple representing the bounds for x.
    """
    
    if _self.MCType == 'ihtmc':
        if _self.function == 'gompertz':
            bounds = (np.log(1E-20),np.log(10))
        elif _self.function == 'exponential':
            bounds = (np.log(1E-20),np.log(50)) 
        else:
            bounds = (np.log(1E-18),np.log(50))
    elif _self.MCType == 'dtmc':
        bounds = (0,1)
        
    return [bounds] * _self.number_parameters()

def bounds_s0(_self) -> list:
    """
    Generates and returns a list of bounds for the initial state values (s0).

    The bounds for each state in 'states' are set to (0, 1000).

    Returns:
        list: A list of tuples representing the bounds for s0, with one tuple per state.
    """
    lims = invert_softmax(_self.s0)
    return [(min(lims),max(lims))] * len(_self.states)

def Transform(_self, param):
    """
    Applies transformations to the input parameters based on the Markov Chain type.

    Parameters:
    _self: Object that contains MCType attribute to determine the type of transformation.
    param: Dictionary containing 'x' and 's0' keys with values to be transformed.

    Raises:
    ValueError: If 'x' or 's0' contains NaN or Inf values after transformation.

    Returns:
    dict: The transformed parameters.
    """
    # Apply exponential transformation to 'x' if MCType is 'ihtmc'
    if _self.MCType == 'ihtmc':
        param['x'] = np.exp(param['x'])
    
    # Apply softmax transformation to 's0'
    param['s0'] = softmax(param['s0'])
    
    # Check for NaN or Inf values in 'x' and 's0'
    if np.isnan(param['s0']).any() or np.isinf(param['s0']).any() or np.isnan(param['x']).any() or np.isinf(param['x']).any():
        raise ValueError(f'There is a NaN or Inf Value in x, s0\n--> x: {param["x"]}\n--> s0: {param["s0"]}')
    
    return param

def InverseTransform(_self, x, s0):
    """
    Applies inverse transformations to 'x' and 's0' based on the Markov Chain type.

    Parameters:
    _self: Object that contains MCType attribute to determine the type of inverse transformation.
    x: Array-like, values to be inverse transformed.
    s0: Array-like, initial state probabilities to be inverse transformed.

    Raises:
    ValueError: If MCType is not recognized.

    Returns:
    np.ndarray: The concatenated inverse transformed 'x' and 's0'.
    """
    # Apply inverse transformation based on MCType
    if _self.MCType == 'ihtmc':
        return np.concatenate((np.log(x), invert_softmax(s0)))
    elif _self.MCType == 'dtmc':
        return np.concatenate((x, invert_softmax(s0)))
    else:
        raise ValueError('The type of Markov chain is not recognized.')

    
    